var fs = require("fs");
var http = require("http");
var util = require("util");


/* *****************************************************************************
 *	FUNCTION:		objTxt
 *	DESCRIPTION:	provide text representation of <o>
 */
 function objTxt(o) {
	 return util.inspect(o,99999,99999);
 }
 
/* *****************************************************************************
 *	FUNCTION:		decodeObfuscated
 *	DESCRIPTION:	decode an obfuscated string
 *					obfuscation is assumed to be like "I can't open that door, Dave"
 */
 function decodeObfuscated(obf) {
	 for (var i=0, res = ""; i<obf.length; i++) {
		res += String.fromCharCode(obf.charCodeAt(i)+1);
	 }
	 return res;
 }

 /* *****************************************************************************
 *	FUNCTION:		encodeObfuscated
 *	DESCRIPTION:	encode an obfuscated string
 *					obfuscation is assumed to be like "I can't open that door, Dave"
 */
 function encodeObfuscated(clr) {
	 for (var i=0, res = ""; i<clr.length; i++) {
		res += String.fromCharCode(clr.charCodeAt(i)-1);
	 }
	 return res;
 }

/* *****************************************************************************
 *	FUNCTION:		fixedEncodeURIComponent
 *	DESCRIPTION:	encodes <str> according to rules in URI-encoding
 *					uses library function encodeURIComponent PLUS
 *					encode !'()* characters in their %xx representation
 */
function fixedEncodeURIComponent (str) {
  return encodeURIComponent(str).replace(/[!'()*]/g, function(c) {
    return '%' + c.charCodeAt(0).toString(16);
  });
}

/* *****************************************************************************
 *	CLASS:			TextOutput
 *	DESCRIPTION:	collects text output in a memory buffer
 */
 function TextOutput() {
	 this.textBuffer = new Array();
	 this.lineBuffer = "";
	 this.lineSeparator = "\n";
	 this.newLine = function(options) {
		 options = options || {};
		 if ((options.only_flush && (this.lineBuffer.length>0)) || (!options.only_flush)) {
			this.textBuffer.push(this.lineBuffer);
			this.lineBuffer = "";
		 }
	 }		 
	 this.print = function() {
		 // console.log("print: have "+arguments.length+" parameters");
		 for (var i=0; i<arguments.length; i++) {
			 if (Array.isArray(arguments[i])) {
				 this.newLine({only_flush: true});
				 this.textBuffer = this.textBuffer.concat(arguments[i]); 
			 } else {
				this.lineBuffer += arguments[i];
			 }
		 }
	 }
	 this.println = function() {
		 this.print.apply(this,arguments);
		 this.newLine();
	 }
	 this.getText = function() {
		 // console.log("getText: have "+this.textBuffer.length+" lines");
		 return this.textBuffer.join(this.lineSeparator)
		 +((this.lineBuffer.length>0)?(this.lineSeparator+this.lineBuffer):"");
	 }
	 // saveToFile:	save the text contents of this instance to a file
	 this.saveToFile = function(filename) {
		 fs.writeFileSync(filename,this.getText(),{encoding:"utf8"});
	 }
	 // clear:	reset this instance to all being empty
	 this.clear = function() {
		 this.textBuffer = new Array();
		 this.lineBuffer = "";
	 }
 }

var embs = {};

/* *****************************************************************************
 *	FUNCTION:		BIN
 */
 function BIN(num,minlen,maxlen) {
	 var s = num.toString(2);
	 while (s.length<maxlen) s = "0"+s;
	 return s;
 }
 
/* *****************************************************************************
 *	FUNCTION:		HEX
 */
 function HEX(num,minlen,maxlen) {
	 var s = num.toString(16);
	 while (s.length<maxlen) s = "0"+s;
	 return s;
 }
 
/* *****************************************************************************
 *	FUNCTION:		readEmbeds
 *	DESCRIPTION:	read embedded wikitext fragments from the source code of this very program
 */
 function readEmbeds() {
	 var sf = fs.readFileSync(__filename,{encoding: "utf8"});	
	 sf = sf.replace(/\r\n/g,"\n"); 	// convert CRLF to LF							
	 sf = sf.replace(/\n\r/g,"\n");		// convert LFCR to LF								
	 sf = sf.replace(/\r/g,"");	        // remove any remaining CRs
	 sflines = sf.split("\n");
	 var clparams, cl, tcl;
	 var vars = {};
	 var state = 'SCAN';
	 for (var li=0; li<sflines.length; li++) {
		 cl = sflines[li];
		 tcl = cl.trim();
		 // console.log("read_embeds.line="+li+".state="+state+".line=\""+cl+"\"");
		 switch (state) {
			 case 'SCAN':
				if (tcl.startsWith("/*")) {
					state = 'COMMENT';
				} else if (cl.startsWith("*/")) {
					state = 'SCAN';
				}
				break;
			case 'COMMENT':
				if (tcl.startsWith("#")) {
					clparams = tcl.split(" ");
					switch (clparams[0]) {
						case "#embed_wikitext":
							vars.tfn = clparams[1];
							vars.tfb = new Array();
							state = 'EMBED_WIKITEXT';
							break;
						case "#embed_c_code":
							vars.tfn = clparams[1];
							vars.tfb = new Array();
							state = 'EMBED_C_CODE';
							break;
						default:
							console.error("Unknown command line "+cl);
					}
				} else if (tcl.startsWith("*/")) {
					state = "SCAN";
				}
				break;
			case 'EMBED_WIKITEXT':
				if (cl.startsWith('#embed_end')) {
					embs[vars.tfn] = vars.tfb;
					state = 'COMMENT';
				} else {
					vars.tfb.push(cl);
				}
				break;
			case 'EMBED_C_CODE':
				if (cl.startsWith('#embed_end')) {
					clparams = tcl.split(" ");
					vars.tfb.push(clparams[1]);
					embs[vars.tfn] = vars.tfb;
					if (clparams[1]=="*/") {
						state = 'SCAN';
					} else {
						state = 'COMMENT';
					}
				} else {
					vars.tfb.push(cl);
				}
				break;
			 default:
				console.error("Unknown state "+state);
		 }
	 }
	 console.log("Got "+Object.keys(embs).length+" embedded fragments");
 }
		  
/* *****************************************************************************
 *	FUNCTION:	mcuPin2bbrPin
 *	INPUT:		mcupin	-	pin number on ATmega6490 mcu chip (TQFP-100)
 *	RESULT:		corresponding pin numer on the rotated TQFP-100 breakout board (bbr)
 *	DESCRIPTION:	When soldering the AVR6490 Chip in the TQFP-100 case to it's breakout board, I made an error:
 *					I turned it 90 degrees to the right. Therefore Pin 1 of the chip appears at Pin 76 of the breakout board
 *					This function provides a correction lookup
 */


function mcuPin2bbrPin(mcupin) {
	return (mcupin + 75) % 100;
}

/* *****************************************************************************
 *	FUNCTION:		genWiring
 *	DESCRIPTION:	generates wikitext for connections from MCU(ATmega6490) to DE-301 LCD module
 */
 function genWiring(wtx) {
	wtx.println("== Connecting D301 LCD to ATmega6490 MCU ==");	
	wtx.println("");
	wtx.println("{{TableBegin}}");
	wtx.println("! LCD pin name !! pin functions !! MCU pin name !! MCU pin number !! BBR pin number");
	var r, fn, cf;
	for (var lpn in pugauge_lcd_pins) {
		r = pugauge_lcd_pins[lpn];
		wtx.println("|-");
		wtx.println("| "+lpn);
		wtx.println("|");
		wtx.println("{{TableBegin}}");
		wtx.println("! function/plane !! digit !! segment");
		for (fn in r) {
			if (!fn.startsWith("com")) continue;
			cf = r[fn];
			wtx.println("|-");
			wtx.println("| "+fn+" || "+cf.digit+" || "+cf.segment);
		}
		wtx.println("{{TableEnd}}");
		wtx.println("| "+r.mcu_pin_name+" || "+mcu_pins[r.mcu_pin_name].mcu_pin+" || "+mcuPin2bbrPin(mcu_pins[r.mcu_pin_name].mcu_pin));
	}
	wtx.println("{{TableEnd}}");
 }

// parameters for drawing segments, segment names an order as in DE-301
var segparams = [
	{	bitnum:	 0,	name:	'seg_a',	start:	{	x:	1,	y:	0	},	end:	{	x:	9,	y:	0	}},
	{	bitnum:	 1,	name:	'seg_h',	start:	{	x:	1,	y:	1	},	end:	{	x:	4,	y:	9	}},
	{	bitnum:	 2,	name:	'seg_x',	start:	{	x:	-2,	y:	-3	},	end:	{	x:	0,	y:	-1	}},
	{	bitnum:	 3,	name:	'seg_i',	start:	{	x:	5,	y:	1	},	end:	{	x:	5,	y:	9	}},
	{	bitnum:	 4,	name:	'seg_b',	start:	{	x:	10,	y:	1	},	end:	{	x:	10,	y:	9	}},
	{	bitnum:	 5,	name:	'seg_g',	start:	{	x:	1,	y:	10	},	end:	{	x:	4,	y:	10	}},
	{	bitnum:	 6,	name:	'seg_f',	start:	{	x:	0,	y:	1	},	end:	{	x:	0,	y:	9	}},
	{	bitnum:	 7,	name:	'seg_j',	start:	{	x:	6,	y:	9	},	end:	{	x:	9,	y:	1	}},
	{	bitnum:	 8,	name:	'seg_c',	start:	{	x:	10,	y:	11	},	end:	{	x:	10,	y:	19	}},
	{	bitnum:	 9,	name:	'seg_l',	start:	{	x:	1,	y:	19	},	end:	{	x:	4,	y:	11	}},
	{	bitnum:	10,	name:	'seg_e',	start:	{	x:	0,	y:	11	},	end:	{	x:	0,	y:	19	}},
	{	bitnum:	11,	name:	'seg_k',	start:	{	x:	6,	y:	10	},	end:	{	x:	9,	y:	10	}},
	{	bitnum:	12,	name:	'seg_dp',	start:	{	x:	11,	y:	21	},	end:	{	x:	11,	y:	21	}},
	{	bitnum:	13,	name:	'seg_m',	start:	{	x:	5,	y:	11	},	end:	{	x:	5,	y:	19	}},
	{	bitnum:	14,	name:	'seg_d',	start:	{	x:	0,	y:	20	},	end:	{	x:	10,	y:	20	}},
	{	bitnum:	15,	name:	'seg_n',	start:	{	x:	6,	y:	11	},	end:	{	x:	9,	y:	19	}}
	];
	
function genSegment(wtx,segnum) {
	var sd = segparams[segnum];
	wtx.println('  <line x1="'+sd.start.x+'" y1="'+sd.start.y+'" x2="'+sd.end.x+'" y2="'+sd.end.y+'" stroke="black" stroke-width="2"/>');
}

function genGlyph(wtx,code) {
	var msk = pugauge_chargen[code];
	for (var bpos=0; bpos<16; bpos++) {
		if (msk&(1<<bpos)) genSegment(wtx,bpos);
	}
}
 
/* *****************************************************************************
 *	FUNCTION:		genFontView
 *	DESCRIPTION:	generates wikitext showing the PUGAUGE font
 */
function genFontView(wtx) {
	wtx.println('== Font made visible ==');
	wtx.println("<html>");
	wtx.println('<svg xmlns="http://www.w3.org/2000/svg" ');
	wtx.println('     xmlns:xlink="http://www.w3.org/1999/xlink" ');
	wtx.println('     version="1.1" ');
	wtx.println('     baseProfile="full" ');
	wtx.println('     width="210mm" height="290mm"');
    wtx.println('     viewBox="-10 -10 800 600">');
	var xpos, ypos;
	for (var cnum=0; cnum<pugauge_chargen.length; cnum++) {
		ypos = 40 * (Math.floor(cnum / 8));
		xpos = 40 * (cnum % 8);
		wtx.println('<g transform="scale(1 1) translate('+xpos+' '+ypos+')">');
		genGlyph(wtx,cnum);
		wtx.println('</g>');
	}
	wtx.println("</svg>");
	wtx.println("</html>");
}

 var wikihost = "scherer3002.duckdns.org";
 var wikiroot = "/public/wiki/"
 var wikipath = wikiroot+"api.php"
 var wikiurl = "http://"+wikihost+wikipath;
 var wikibaseurl = "http://"+wikihost+wikiroot+"index.php/";
 
/* *****************************************************************************
 *	FUNCTION:		doWikiApi
 *	DESCRIPTION:	perform a mediawiki GET API call
 */
function doWikiApi(grq,txt,rfunc) {
	http.get(wikiurl+"?"+grq, 
	         (resp) => {
				var data = '';
				var resobj;
				// A chunk of data has been received.
				resp.on('data', (chunk) => {
								data += chunk;
							});
				// The whole response has been received. Print out the result.
				resp.on('end', 
						() => {
							try {
								resobj = JSON.parse(data);
							} catch(e) {
								resobj = {error: e, data: data};
							}
							console.log('doWikiApi: RES('+(data.length)+'): '+util.inspect(resobj,99999,99999));
							rfunc(resobj);
							});
				resp.on("error", 
						(err) => {
							console.log("doWikiApi: Error: " + err.message);
							});
			});
}

/* *****************************************************************************
 *	FUNCTION:		doWikiPostApi
 *	DESCRIPTION:	perform a mediawiki POST API call
 */
function doWikiPostApi(rq,rfunc) {
	var grq = rq.req || "";
	var txt = rq.content || "";
	var cookie = rq.cookie || "";
	if (txt.length>0) {
		txt = grq+"\n"+txt;
	} else {
		txt = grq;
	}
	// console.log("doWikiPostApi: TXT(len="+Buffer.byteLength(txt)+"): "+txt);
	var options = {
		hostname:	wikihost,
		port: 		80,
		path:		wikipath,
		method:		'POST',
		headers:	{
						'Content-Type': 'application/x-www-form-urlencoded',
						'Content-Length': Buffer.byteLength(txt)
					}
		};
	if (cookie) options.headers['Cookie'] = cookie;
	// console.log("doWikiPostApi: options: "+objTxt(options));
	var req = http.request(options,
				(res) => {
					var data = '';
					var resobj;
					var res_status = res.statusCode;
					var res_headers = res.headers;
					var res_cookies = setCookie.parse(res, {
									decodeValues: true  // default: true
									});
					// console.log(`doWikiPostApi: result: STATUS: ${res.statusCode}`);
					// console.log(`doWikiPostApi: result: HEADERS: ${JSON.stringify(res.headers)}`);
					res.setEncoding('utf8');
					res.on('data', (chunk) => {
							// console.log(`doWikiPostApi: BODY: ${chunk}`);
							data += chunk;
						});
					res.on('end', () => {
						try {
							resobj = JSON.parse(data);
							resobj.headers = res_headers;
							resobj.status = res_status;
							resobj.cookies = res_cookies;
						} catch(e) {
							resobj = {error: e, data: data};
						}
						// console.log('doWikiPostApi: RES('+(data.length)+'): '+util.inspect(resobj,99999,99999));
						rfunc(resobj);
						})
					});
	req.on('error', (e) => {
			console.log(`doWikiPostApi: problem with request: ${e.message}`);
		});
	req.write(txt);
	req.end();
}


/* *****************************************************************************
 *	FUNCTION:		
 */
function postTextOutputToWiki(txo,pagetitle) {
	var wikiusers = [{ username: "Projapi", password: "30Dec2K181"}];
	var wikiuser = "Wolfgang"
	var wikipassword = decodeObfuscated(OBF_WIKI_PASSWD);
	var loginreq = "action=login&format=json&lgname="+wikiuser+"&lgpassword="+wikipassword;
	var mwcookie;
	// 1: get login token
	doWikiPostApi({req: loginreq},
		(res) => {
			// console.log("postTextOutputToWiki: initial login result: "+objTxt(res));
			if (res.error || res.warnings) { console.error("postTextOutputToWiki: Error or warnings getting login token: "+objTxt(res)); return; }
			if (!res.login && !res.login.token) { console.error("postTextOutputToWiki: no login token: "+objTxt(res)); return; }
			if (!res.login && !res.login.cookieprefix) { console.error("postTextOutputToWiki: no cookie prefix: "+objTxt(res)); return; }
			var cp = res.login.cookieprefix;
			console.log("postTextOutputToWiki: initial login, cookieprefix: "+cp);
			mwcookie = "";
			var cc;
			for (var ci=0; ci<res.cookies.length; ci++) {
				cc = res.cookies[ci];
				// console.log("postTextOutputToWiki: cookie # "+ci+": "+objTxt(cc));
				if (cc.name.startsWith(cp)) {
					if (mwcookie.length>0) cv += " ";
					mwcookie += (cc.name+"="+cc.value+";");
				}
				// console.log("postTextOutputToWiki: got login key, mwcookie: "+mwcookie);
			}
			// 2: login using login token
			doWikiPostApi({req: loginreq+"&lgtoken="+res.login.token, cookie: mwcookie},
				(res) => {
					if (res.error || res.warnings) { console.error("postTextOutputToWiki: Error or warnings logging in: "+objTxt(res)); return; }
					// console.log("postTextOutputToWiki: login information: "+objTxt(res));
					// 3: get edit token
						doWikiPostApi({req: "action=tokens&format=json", cookie: mwcookie},
						(res) => {
							if (res.error || res.warnings) { console.error("postTextOutputToWiki: Error or warnings in getting edit token: "+objTxt(res)); return; }
							if (!res.tokens && !res.tokens.edittoken) { console.error("postTextOutputToWiki: no edit token: "+objTxt(res)); return; }
							// console.log("postTextOutputToWiki: response with edit token: "+objTxt(res));
							var txt = txo.getText();
							// console.log("postTextOutputToWiki: text has "+txt.length+" characters");
							txt = fixedEncodeURIComponent(txt);
							// console.log("postTextOutputToWiki: URL-encoded text has "+txt.length+" characters");
							var edittoken = res.tokens.edittoken;
							// 4: edit page using edit token
							doWikiPostApi({req: "action=edit&format=json&token="+fixedEncodeURIComponent(edittoken)+
												"&title="+pagetitle+"&text="+txt, 
												/* content: txo.getText(), */ 
												cookie: mwcookie},
								(res) => {
									if (res.error || res.warnings) { console.error("postTextOutputToWiki: Error or warnings editing page: "+objTxt(res)); return; }
									console.log("postTextOutputToWiki: posting wiki-page "+pagetitle+" complete: Page-ID: "+res.edit.pageid+", page: "+wikibaseurl+(res.edit.title.replace(" ","_")));
								});
						});
				});
		});
}

/* *****************************************************************************
 *	FUNCTION:		genFileFromTemplate
 *	INPUT:			ifn			-	template file name
 *					rdt			-	replacement data
 *					ofn			-	output file name
 */
 function genFileFromTemplate(ifn,rdt,ofn) {
	var ttx = fs.readFileSync(ifn,{encoding:'utf8'});
	var otx = ttx.replace(/\%\%([A-Za-z0-9]+)\%\%/g,
							function(match,mstr,pos) {
								// console.log("Found match "+match+" match "+mstr+" at position "+pos);
								if (rdt.hasOwnProperty(mstr)) {
									return rdt[mstr];
								} else {
									return "<undefined."+mstr+">";
								}
							});
	fs.writeFileSync(ofn,otx,{encoding: "utf8"});
 }
 
/* *****************************************************************************
 *	FUNCTION:		
 */
function genArduinoPlatformZip(droot,idt) {
	const child_process = require('child_process');
	const sfrs = [
		{ 	field:	"CHECKSUM"	},
		{	field:	"SIZE"		},
		{	field:	"VERSION"	},
		{	field:	"FILENAME"	},
		{	field:	"STATUS"	}
	];
	const csfn = "./CHECKSUM_SIZE.txt";
	var stdout = child_process.execFileSync('bash', ["./pack.release.bash"], {shell: true});
	var stx = fs.readFileSync(csfn,{encoding: 'utf8'});
	stx = stx.replace(/\r\n/g,"\n");
	stx = stx.replace(/\n\r/g,"\n");
	stx = stx.replace(/\r/g,"");
	var slns = stx.split("\n");
	var sl;
	var slfs;
	if (slns.length<3) {
		console.error("status file "+csfn+" should have at least "+sfrs.length+" lines, but has "+slns.length);
		return;
	}
	for (var li=0; li<sfrs.length; li++) {
		sl = slns[li];
		slfs = sl.split(/\s+/g);
		if (slfs.length!=2) {
			console.error("status file "+csfn+" line "+(li+1)+" should have 2 fields, but has "+slfs.length); 
			return;
		}
		idt[sfrs[li].field] = slfs[0];
	}
	console.log("package built, result: "+objTxt(idt));
}

/* *****************************************************************************
 *	FUNCTION:		
 */
function genArduinoPackage(droot) {
	var infodat = {};
	genArduinoPlatformZip(droot+"/1.0.0",infodat);
	genFileFromTemplate(droot+"/1.0.0/extras/package_index.json.template",
						infodat,
						droot+"/../downloads/package_pucon_pugauge_index.json");

}
 
/* *****************************************************************************
 *				M A I N   P R O G R A M
 * *****************************************************************************/
 
var wtx = new TextOutput();
readEmbeds();
genArduinoPackage("/mnt/puconnas01/public/Projects/PuProjects/PUGAUGE/ArduinoCore-pucon_pugauge");